
## 2 RDF and Types

### 2.1 Triples
The group of standards and ecosystem surrounding `OWL` and `RDF` represent the current state of semantic web technologies. 

Both of these standards are triple based, i.e. the fundamental unit is the triple `subject-predicate-object`.

***Postulate***: The triple formulation is very convenient unit for low level use, but not for high level definitions. 

The standard tool for working with OWL and RDF is [protege](https://protege.stanford.edu/) which illustrates the point of the [context](#11-context) on complexity as a limiting factor.

***Postulate***: An approach that imports from `OWL` and interoperates wth `RDF` but focuses on static types is required for both simplicity and use potential




### 2.2 JSON-LD @type

Fortunately, [JSON-LD](https://json-ld.org/) is an RDF [standard](https://w3c.github.io/json-ld-syntax/) that defines a type (`@type`) as the set of a subject's triples. A good introduction to role of JSON-LD and the web is given [here](http://www.markus-lanthaler.com/research/on-using-json-ld-to-create-evolvable-restful-services.pdf)

When we equate the static types in our programming language with those in the ontology, we are practically speaking, making this mapping via `JSON-LD` types. 

Programming types are marshalled to `JSON-LD` (stringified) and unmarshalled (loaded) from `JSON-LD` to instances of static types in a program.

### 2.3 JSON-LD Schema

A limitation of `JSON-LD`is that there is no JSON-LD Schema standard.  The standard does not define a means of defining the pure type (schema) and we have to do this via some other mechanism. 

We are currently defining the type definitions via annotated type (schema) definitions made in `JSON-LD` as per [TerminusDB](https://terminusdb.com/docs/schema-reference-guide/).

In practice, `JSON-LD` is used for semantic data serialization, interface types, and database storage.  

Pure schema or type definitions have a number of uses. The schema definition are typically compile time artifacts available via libaries to both client and server without access to the database and can be used for a number of purposes:
  - program languge type defintions
  - instance validation
  - code generation
  - reasoning about types
   
All of the following ontologies are generated from definitions of the types in a specific cluster.

### 2.4 Visualization

The JSON-LD based schema definitions are used to generate graphical representation of ontology clusters as per the following `core` cluster.
The outer scope defines the schema and version where the cluster is used. The cluster scoping can be added or removed so that multiple custers, or a complete schema, can be shown together. It is simpler to view clusters individually and show dependencies (also a graph option) on exteral cluster types. The predicate names can optionally be repeated on the edges defining the association but this does contribute information as the Object's type is listed together with the predicate name.


| Graph key | interpretation |
|:--- |:--- |
| type name | box heading is value of @type |
| colored type box| abstract type |
| black type box | concrete type |
| predicate name listed on left | object type or literal type on right |
| dotted arrow |  subsumes |
| solid arrow | predicate |


![core](../ontologies/clusters/core/facilityCobot-v022-core-cluster.svg)

### 2.5 Dimensions

Note that the underlying dimesions are defined in `core`cluster along with the common base type `CommonEntity`:
  - `AbstractEntity`
  - `SampledEntity`and `TemporalEntity`
  - `PhysicalEntity`
  - `CyberPhysicalActor` (actor)
  - `ActionEntity`

  as per the FFU ontology derived from [IHMC](https://ontology.ihmc.us/ontology.html)

These same dimensions can be found in [DUL](#52-dul) and replace the role of `core`in the FERA schema.

<a id="required-core"></a>

### 2.6 Required Core properties

***Proposition*** all `URI`s that are exposed to users or programmers should be logically constructed from ***well-known-identifiers*** when ever possible. 

A `URI` is built from a root domain of the service hosting the data and schema instances so the uniqueness requirement typically pertains on a type basis, not global scope basis.

All of the following ontologies are dependent on the use of a unique `name` value for a given type as the lexical URI generation scheme always involves the `type/name` pair. 

This is generally suffient for meta-data but requires an extra `index`, `count`, `timestamp` or some other property value to disambiguate sampled data. 

The type name is implied by default, and `name` shall be included in a root type of the ontology.

Just as the design of `URL`s is an important and overlooked part of the design of `RESTful`services, the same consideration on the lexical design of `URI`s should be made in conjunction with the type design.

***Postulate*** the ontology designer should try to keep all predicates (property keys) unique to the type

This means that if you want to reuse the `name`, or any other predicate in a new type, you should inherit the type that it is defined in, rather than duplicating it in your new type. 

Note that handling of predicate naming may differ between a type centric and a triple centric approach.


### 2.7 URIs

All subjects, predicates, and objects are identified by URIs (LinkedData) and are generated by the database according to a lexical scheme defined per type in the schema definition.

One of the practial programming challenges of working with graphs of semantic data is ensuring that you have the URIs for all the neighboring objects that you are linking to. 

Use of a well defined lexical `URI`generation scheme means that users can emulate a object reference without a database roundtrip.

The issue of providing parent and neighboring `URI`s in creating a semantic data instance is simplified for the user by the use of type constructors that should be bundled in a cluster's stub libary.

A link failure means that the object does not exist (closed world assumption), not that this object may exist but that we just do not know about it (open world assumption).



[HOME](../../README.md)